# HG changeset patch
# Parent 4fb76c06a6a8ec42de9381ecec13663f437379fe

diff --git a/js/src/builtin/SymbolObject.cpp b/js/src/builtin/SymbolObject.cpp
--- a/js/src/builtin/SymbolObject.cpp
+++ b/js/src/builtin/SymbolObject.cpp
@@ -44,6 +44,7 @@ const JSFunctionSpec SymbolObject::metho
 };
 
 const JSFunctionSpec SymbolObject::staticMethods[] = {
+    JS_FN("for", for_, 1, 0),
     JS_FS_END
 };
 
@@ -102,6 +103,25 @@ SymbolObject::construct(JSContext *cx, u
     return true;
 }
 
+// ES6 rev 24 (2014 Apr 27) 19.4.2.2
+bool
+SymbolObject::for_(JSContext *cx, unsigned argc, Value *vp)
+{
+    CallArgs args = CallArgsFromVp(argc, vp);
+
+    // steps 1-2
+    RootedString stringKey(cx, ToString(cx, args.get(0)));
+    if (!stringKey)
+        return false;
+
+    // steps 3-7
+    JS::Symbol *symbol = JS::Symbol::forNonAtom(cx, stringKey);
+    if (!symbol)
+        return false;
+    vp->setSymbol(symbol);
+    return true;
+}
+
 JSObject *
 js_InitSymbolClass(JSContext *cx, HandleObject obj)
 {
diff --git a/js/src/builtin/SymbolObject.h b/js/src/builtin/SymbolObject.h
--- a/js/src/builtin/SymbolObject.h
+++ b/js/src/builtin/SymbolObject.h
@@ -42,6 +42,9 @@ class SymbolObject : public JSObject
 
     static bool construct(JSContext *cx, unsigned argc, Value *vp);
 
+    // Static methods.
+    static bool for_(JSContext *cx, unsigned argc, Value *vp);
+
     static const JSPropertySpec properties[];
     static const JSFunctionSpec methods[];
     static const JSFunctionSpec staticMethods[];
diff --git a/js/src/gc/Statistics.cpp b/js/src/gc/Statistics.cpp
--- a/js/src/gc/Statistics.cpp
+++ b/js/src/gc/Statistics.cpp
@@ -288,6 +288,7 @@ static const PhaseInfo phases[] = {
     { PHASE_SWEEP_MARK_GRAY_WEAK, "Mark Gray and Weak", PHASE_SWEEP_MARK },
     { PHASE_FINALIZE_START, "Finalize Start Callback", PHASE_SWEEP },
     { PHASE_SWEEP_ATOMS, "Sweep Atoms", PHASE_SWEEP },
+    { PHASE_SWEEP_SYMBOL_REGISTRY, "Sweep Symbol Registry", PHASE_SWEEP },
     { PHASE_SWEEP_COMPARTMENTS, "Sweep Compartments", PHASE_SWEEP },
     { PHASE_SWEEP_DISCARD_CODE, "Sweep Discard Code", PHASE_SWEEP_COMPARTMENTS },
     { PHASE_SWEEP_TABLES, "Sweep Tables", PHASE_SWEEP_COMPARTMENTS },
diff --git a/js/src/gc/Statistics.h b/js/src/gc/Statistics.h
--- a/js/src/gc/Statistics.h
+++ b/js/src/gc/Statistics.h
@@ -39,6 +39,7 @@ enum Phase {
     PHASE_SWEEP_MARK_GRAY_WEAK,
     PHASE_FINALIZE_START,
     PHASE_SWEEP_ATOMS,
+    PHASE_SWEEP_SYMBOL_REGISTRY,
     PHASE_SWEEP_COMPARTMENTS,
     PHASE_SWEEP_DISCARD_CODE,
     PHASE_SWEEP_TABLES,
diff --git a/js/src/jsapi-tests/testSymbol.cpp b/js/src/jsapi-tests/testSymbol.cpp
--- a/js/src/jsapi-tests/testSymbol.cpp
+++ b/js/src/jsapi-tests/testSymbol.cpp
@@ -24,3 +24,30 @@ BEGIN_TEST(testSymbol_New)
     return true;
 }
 END_TEST(testSymbol_New)
+
+BEGIN_TEST(testSymbol_GetSymbolFor)
+{
+    RootedString desc(cx, JS_NewStringCopyZ(cx, "ponies"));
+    CHECK(desc);
+    Rooted<Symbol*> sym1(cx);
+    CHECK(sym1 = GetSymbolFor(cx, desc));
+    CHECK_SAME(StringValue(GetSymbolDescription(sym1)), StringValue(desc));
+
+    // Calling JS::GetSymbolFor again with the same arguments produces the
+    // same Symbol.
+    Rooted<Symbol*> sym2(cx);
+    CHECK(sym2 = GetSymbolFor(cx, desc));
+    CHECK_EQUAL(sym1, sym2);
+
+    // Passing a new but equal string also produces the same Symbol.
+    CHECK(desc = JS_NewStringCopyZ(cx, "ponies"));
+    CHECK(sym2 = GetSymbolFor(cx, desc));
+    CHECK_EQUAL(sym1, sym2);
+
+    // But SymbolNew always produces a new distinct Symbol.
+    CHECK(sym2 = NewSymbol(cx, desc));
+    CHECK(sym2 != sym1);
+
+    return true;
+}
+END_TEST(testSymbol_GetSymbolFor)
diff --git a/js/src/jsapi.cpp b/js/src/jsapi.cpp
--- a/js/src/jsapi.cpp
+++ b/js/src/jsapi.cpp
@@ -5687,6 +5687,16 @@ JS::NewSymbol(JSContext *cx, HandleStrin
     return Symbol::new_(cx, description);
 }
 
+JS_PUBLIC_API(JS::Symbol *)
+JS::GetSymbolFor(JSContext *cx, HandleString key)
+{
+    AssertHeapIsIdle(cx);
+    CHECK_REQUEST(cx);
+    assertSameCompartment(cx, key);
+
+    return Symbol::forNonAtom(cx, key);
+}
+
 JS_PUBLIC_API(JSString *)
 JS::GetSymbolDescription(Handle<Symbol*> symbol)
 {
diff --git a/js/src/jsapi.h b/js/src/jsapi.h
--- a/js/src/jsapi.h
+++ b/js/src/jsapi.h
@@ -4361,6 +4361,16 @@ JS_PUBLIC_API(Symbol *)
 NewSymbol(JSContext *cx, HandleString description);
 
 /*
+ * Symbol.for as specified in ES6.
+ *
+ * Get a Symbol with the description 'key' from the Runtime-wide symbol registry.
+ * If there is not already a Symbol with that description in the registry, a new
+ * Symbol is created and registered. 'key' must not be null.
+ */
+JS_PUBLIC_API(Symbol *)
+GetSymbolFor(JSContext *cx, HandleString key);
+
+/*
  * Get the [[Description]] attribute of the given symbol.
  *
  * This function is infallible. If it returns null, that means the symbol's
diff --git a/js/src/jscntxt.h b/js/src/jscntxt.h
--- a/js/src/jscntxt.h
+++ b/js/src/jscntxt.h
@@ -388,6 +388,9 @@ class ExclusiveContext : public ThreadSa
     JSCompartment *atomsCompartment() {
         return runtime_->atomsCompartment();
     }
+    SymbolRegistry &symbolRegistry() {
+        return runtime_->symbolRegistry();
+    }
     ScriptDataTable &scriptDataTable() {
         return runtime_->scriptDataTable();
     }
diff --git a/js/src/jsgc.cpp b/js/src/jsgc.cpp
--- a/js/src/jsgc.cpp
+++ b/js/src/jsgc.cpp
@@ -3880,8 +3880,14 @@ GCRuntime::beginSweepingZoneGroup()
     }
 
     if (sweepingAtoms) {
-        gcstats::AutoPhase ap(stats, gcstats::PHASE_SWEEP_ATOMS);
-        rt->sweepAtoms();
+        {
+            gcstats::AutoPhase ap(stats, gcstats::PHASE_SWEEP_ATOMS);
+            rt->sweepAtoms();
+        }
+        {
+            gcstats::AutoPhase ap(stats, gcstats::PHASE_SWEEP_SYMBOL_REGISTRY);
+            rt->symbolRegistry().sweep();
+        }
     }
 
     /* Prune out dead views from ArrayBuffer's view lists. */
diff --git a/js/src/tests/ecma_6/Map/symbol-as-key.js b/js/src/tests/ecma_6/Map/symbol-as-key.js
--- a/js/src/tests/ecma_6/Map/symbol-as-key.js
+++ b/js/src/tests/ecma_6/Map/symbol-as-key.js
@@ -17,5 +17,17 @@ m.delete(sym);
 assertEq(m.has(sym), false);
 assertEq(m.size, 0);
 
+// Symbols returned by Symbol.for() can be Map keys.
+for (var word of "that that is is that that is not is not is that not it".split(' ')) {
+    sym = Symbol.for(word);
+    m.set(sym, (m.get(sym) || 0) + 1);
+}
+assertDeepEq([...m], [
+    [Symbol.for("that"), 5],
+    [Symbol.for("is"), 5],
+    [Symbol.for("not"), 3],
+    [Symbol.for("it"), 1]
+]);
+
 if (typeof reportCompare === "function")
   reportCompare(0, 0);
diff --git a/js/src/tests/ecma_6/Symbol/for.js b/js/src/tests/ecma_6/Symbol/for.js
new file mode 100644
--- /dev/null
+++ b/js/src/tests/ecma_6/Symbol/for.js
@@ -0,0 +1,31 @@
+/* Any copyright is dedicated to the Public Domain.
+ * http://creativecommons.org/licenses/publicdomain/ */
+
+// Symbol.for called twice with the same argument returns the same symbol.
+assertEq(Symbol.for("ponies"), Symbol.for("ponies"));
+
+// Called twice with equal strings: still the same result.
+var one = Array(64+1).join("x");
+var two = Array(8+1).join(Array(8+1).join("x"));
+assertEq(Symbol.for(one), Symbol.for(two));
+
+// Symbols created by calling Symbol() are not in the symbol registry.
+var sym = Symbol("123");
+assertEq(Symbol.for("123") !== sym, true);
+
+// Empty string is fine.
+assertEq(typeof Symbol.for(""), "symbol");
+
+// Primitive arguments.
+assertEq(Symbol.for(3), Symbol.for("3"));
+assertEq(Symbol.for(null), Symbol.for("null"));
+assertEq(Symbol.for(undefined), Symbol.for("undefined"));
+assertEq(Symbol.for(), Symbol.for("undefined"));
+
+// Symbol.for ignores the 'this' value.
+var foo = Symbol.for("foo")
+assertEq(Symbol.for.call(String, "foo"), foo);
+assertEq(Symbol.for.call(3.14, "foo"), foo);
+
+if (typeof reportCompare === "function")
+    reportCompare(0, 0);
diff --git a/js/src/tests/ecma_6/Symbol/realms.js b/js/src/tests/ecma_6/Symbol/realms.js
--- a/js/src/tests/ecma_6/Symbol/realms.js
+++ b/js/src/tests/ecma_6/Symbol/realms.js
@@ -18,6 +18,10 @@ if (typeof newGlobal === "function") {
     var smith = Symbol("smith");
     g.smith = smith;  // put smith into the realm
     assertEq(g.smith, smith);  // pull it back out
+
+    // Symbol.for functions share a symbol registry across all realms.
+    assertEq(g.Symbol.for("ponies"), Symbol.for("ponies"));
+    assertEq(g.eval("Symbol.for('rainbows')"), Symbol.for("rainbows"));
 }
 
 if (typeof reportCompare === "function")
diff --git a/js/src/tests/ecma_6/Symbol/surfaces.js b/js/src/tests/ecma_6/Symbol/surfaces.js
--- a/js/src/tests/ecma_6/Symbol/surfaces.js
+++ b/js/src/tests/ecma_6/Symbol/surfaces.js
@@ -21,5 +21,7 @@ assertEq(desc.configurable, true);
 assertEq(desc.enumerable, false);
 assertEq(desc.writable, true);
 
+assertEq(Symbol.for.length, 1);
+
 if (typeof reportCompare === "function")
     reportCompare(0, 0);
diff --git a/js/src/tests/ecma_6/Symbol/typeof.js b/js/src/tests/ecma_6/Symbol/typeof.js
--- a/js/src/tests/ecma_6/Symbol/typeof.js
+++ b/js/src/tests/ecma_6/Symbol/typeof.js
@@ -3,6 +3,7 @@
 
 assertEq(typeof Symbol(), "symbol");
 assertEq(typeof Symbol("ponies"), "symbol");
+assertEq(typeof Symbol.for("ponies"), "symbol");
 
 if (typeof reportCompare === "function")
     reportCompare(0, 0);
diff --git a/js/src/vm/Runtime.cpp b/js/src/vm/Runtime.cpp
--- a/js/src/vm/Runtime.cpp
+++ b/js/src/vm/Runtime.cpp
@@ -312,6 +312,9 @@ JSRuntime::init(uint32_t maxbytes)
     atomsZone.forget();
     this->atomsCompartment_ = atomsCompartment.forget();
 
+    if (!symbolRegistry_.init())
+        return false;
+
     if (!scriptDataTable_.init())
         return false;
 
diff --git a/js/src/vm/Runtime.h b/js/src/vm/Runtime.h
--- a/js/src/vm/Runtime.h
+++ b/js/src/vm/Runtime.h
@@ -41,6 +41,7 @@
 #include "vm/MallocProvider.h"
 #include "vm/SPSProfiler.h"
 #include "vm/Stack.h"
+#include "vm/Symbol.h"
 #include "vm/ThreadPool.h"
 
 #ifdef _MSC_VER
@@ -1175,15 +1176,21 @@ struct JSRuntime : public JS::shadow::Ru
 
   private:
     // Set of all atoms other than those in permanentAtoms and staticStrings.
-    // This may be modified by threads with an ExclusiveContext and requires
-    // a lock.
+    // Reading or writing this set requires the calling thread to have an
+    // ExclusiveContext and hold a lock. Use AutoLockForExclusiveAccess.
     js::AtomSet *atoms_;
 
-    // Compartment and associated zone containing all atoms in the runtime,
-    // as well as runtime wide IonCode stubs. The contents of this compartment
-    // may be modified by threads with an ExclusiveContext and requires a lock.
+    // Compartment and associated zone containing all atoms in the runtime, as
+    // well as runtime wide IonCode stubs. Modifying the contents of this
+    // compartment requires the calling thread to have an ExclusiveContext and
+    // hold a lock. Use AutoLockForExclusiveAccess.
     JSCompartment *atomsCompartment_;
 
+    // Set of all live symbols produced by Symbol.for(). All such symbols are
+    // allocated in the atomsCompartment. This table is subject to the same
+    // locking scheme as atoms_ and atomsCompartment_.
+    js::SymbolRegistry symbolRegistry_;
+
   public:
     bool initializeAtoms(JSContext *cx);
     void finishAtoms();
@@ -1208,6 +1215,11 @@ struct JSRuntime : public JS::shadow::Ru
 
     bool activeGCInAtomsZone();
 
+    js::SymbolRegistry &symbolRegistry() {
+        JS_ASSERT(currentThreadHasExclusiveAccess());
+        return symbolRegistry_;
+    }
+
     // Permanent atoms are fixed during initialization of the runtime and are
     // not modified or collected until the runtime is destroyed. These may be
     // shared with another, longer living runtime through |parentRuntime| and
diff --git a/js/src/vm/Symbol.cpp b/js/src/vm/Symbol.cpp
--- a/js/src/vm/Symbol.cpp
+++ b/js/src/vm/Symbol.cpp
@@ -10,7 +10,22 @@ using JS::Symbol;
 using namespace js;
 
 Symbol *
-Symbol::new_(ExclusiveContext *cx, JSString *desc)
+Symbol::newInternal(ExclusiveContext *cx, JSAtom *desc, bool inRegistry)
+{
+    MOZ_ASSERT(cx->compartment() == cx->atomsCompartment());
+    MOZ_ASSERT(cx->atomsCompartment()->runtimeFromAnyThread()->currentThreadHasExclusiveAccess());
+
+    // Following js::AtomizeString, we grudgingly forgo last-ditch GC here.
+    Symbol *p = gc::AllocateNonObject<Symbol, NoGC>(cx);
+    if (!p) {
+        js_ReportOutOfMemory(cx);
+        return nullptr;
+    }
+    return new (p) Symbol(desc, inRegistry);
+}
+
+Symbol *
+Symbol::new_(ExclusiveContext *cx, JSString *desc, bool inRegistry)
 {
     RootedAtom atom(cx);
     if (desc) {
@@ -21,12 +36,50 @@ Symbol::new_(ExclusiveContext *cx, JSStr
 
     AutoLockForExclusiveAccess lock(cx);
     AutoCompartment ac(cx, cx->atomsCompartment());
+    return newInternal(cx, atom, inRegistry);
+}
 
-    // Following js::AtomizeString, we grudgingly forgo last-ditch GC here.
-    Symbol *p = gc::AllocateNonObject<Symbol, NoGC>(cx);
-    if (!p) {
+Symbol *
+Symbol::for_(js::ExclusiveContext *cx, JSAtom *desc)
+{
+    MOZ_ASSERT(desc);
+    AutoLockForExclusiveAccess lock(cx);
+
+    SymbolRegistry &registry = cx->symbolRegistry();
+    SymbolRegistry::AddPtr p = registry.lookupForAdd(desc);
+    if (p)
+        return *p;
+
+    AutoCompartment ac(cx, cx->atomsCompartment());
+    Symbol *sym = newInternal(cx, desc, true);
+    if (!sym)
+        return nullptr;
+
+    // p is still valid here because we have held the lock since the
+    // lookupForAdd call, and newInternal can't GC.
+    if (!registry.add(p, sym)) {
+        // SystemAllocPolicy does not report OOM.
         js_ReportOutOfMemory(cx);
         return nullptr;
     }
-    return new (p) Symbol(atom);
+    return sym;
 }
+
+Symbol *
+Symbol::forNonAtom(js::ExclusiveContext *cx, HandleString desc)
+{
+    JSAtom *atom = AtomizeString(cx, desc);
+    if (!atom)
+        return nullptr;
+    return for_(cx, atom);
+}
+
+void
+SymbolRegistry::sweep()
+{
+    for (Enum e(*this); !e.empty(); e.popFront()) {
+        Symbol *sym = e.front();
+        if (IsSymbolAboutToBeFinalized(&sym))
+            e.removeFront();
+    }
+}
diff --git a/js/src/vm/Symbol.h b/js/src/vm/Symbol.h
--- a/js/src/vm/Symbol.h
+++ b/js/src/vm/Symbol.h
@@ -12,6 +12,8 @@
 #include "js/RootingAPI.h"
 #include "js/TypeDecls.h"
 
+#include "jsalloc.h"
+
 #include "gc/Barrier.h"
 
 namespace JS {
@@ -21,18 +23,33 @@ class Symbol : public js::gc::BarrieredC
   private:
     JSAtom *description_;
     void *unused1_;
-    uint32_t unused2_;
+    bool inSymbolRegistry_;
 
-    explicit Symbol(JSAtom *desc) : description_(desc) {}
+    // Don't worry about the use of bool here. A later patch in the stack
+    // changes it to an enum.
+    Symbol(JSAtom *desc, bool inRegistry)
+        : description_(desc), inSymbolRegistry_(inRegistry) {}
+
     Symbol(const Symbol &) MOZ_DELETE;
     void operator=(const Symbol &) MOZ_DELETE;
 
+    static Symbol *
+    newInternal(js::ExclusiveContext *cx, JSAtom *desc, bool inRegistry);
+
   public:
     static Symbol *
-    new_(js::ExclusiveContext *cx, JSString *desc=nullptr);
+    new_(js::ExclusiveContext *cx, JSString *desc=nullptr, bool inRegistry=false);
+
+    static Symbol *
+    for_(js::ExclusiveContext *cx, JSAtom *desc);
+
+    static Symbol *
+    forNonAtom(js::ExclusiveContext *cx, js::HandleString desc);
 
     JSAtom *description() { return description_; }
 
+    bool isInSymbolRegistry() const { return inSymbolRegistry_; }
+
     static inline js::ThingRootKind rootKind() { return js::THING_ROOT_SYMBOL; }
     inline void markDescription(JSTracer *trc);
     inline void finalize(js::FreeOp *) {}
@@ -40,4 +57,28 @@ class Symbol : public js::gc::BarrieredC
 
 } /* namespace JS */
 
+namespace js {
+
+struct HashSymbolsByDescription
+{
+    typedef JS::Symbol *Key;
+    typedef JSAtom *Lookup;
+
+    static HashNumber hash(Lookup l) {
+        return HashNumber(reinterpret_cast<uintptr_t>(l));
+    }
+    static bool match(Key sym, Lookup l) {
+        return sym->description() == l;
+    }
+};
+
+class SymbolRegistry : public HashSet<JS::Symbol *, HashSymbolsByDescription, SystemAllocPolicy>
+{
+  public:
+    SymbolRegistry() : HashSet() {}
+    void sweep();
+};
+
+} /* namespace js */
+
 #endif /* vm_Symbol_h */
